@@ -2,10 +2,10 @@
 polyanalyst6api.exceptions
 ~~~~~~~~~~~~~~~~~~~~~~~~~~
 
-This module contains polyanlyst6api specific Exception classes.
+This module contains polyanalyst6api specific Exception classes.
 """
 
-__all__ = ['PAException', 'ClientException', 'APIException']
+__all__ = ['PAException', 'ClientException', 'APIException', 'PABusy']
 
 
 class PAException(Exception):
@@ -29,6 +29,14 @@ class APIException(PAException):
         return f'{self.message} ({self.status_code}, {self.endpoint})'
 
 
+class PABusy(APIException):
+    def __init__(self):
+        super().__init__('')
+
+    def __str__(self):
+        return 'PolyAnalyst server is busy'
+
+
 class ClientException(PAException):
     """Indicate errors that don't involve interaction with PolyAnalyst's API."""
 
