@@ -7,12 +7,13 @@ This module contains functionality for access to PolyAnalyst Analytical Client A
 import datetime
 import functools
 import math
+import os
 import time
 import warnings
 from urllib.parse import urlparse, parse_qs
 from typing import Any, Dict, List, Union, Optional, Tuple, Iterator
 
-from .exceptions import APIException, _WrapperNotFound
+from .exceptions import APIException, ClientException, _WrapperNotFound, PABusy
 
 __all__ = ['Project', 'Parameters', 'DataSet']
 
@@ -28,6 +29,9 @@ class Project:
 
     :param api: An instance of :class:`API <API>` class
     :param uuid: The uuid of the project you want to interact with
+
+    .. versionchanged:: 0.31.0
+           Added nodes list updating while initializing the class
     """
 
     def __repr__(self):
@@ -36,18 +40,21 @@ class Project:
     def __init__(self, api, uuid: str):
         self.api = api
         self.uuid = uuid
-        self._node_list: List[Node] = []
+        self._node_list: List[Node] = self.get_node_list() # automatically loads a project
 
     def get_node_list(self) -> List[Node]:
         """Returns a list of project nodes.
 
         .. versionadded:: 0.15.0
         """
-        return self.api.get(
-            'project/nodes',
-            params={'prjUUID': self.uuid},
-            headers={'sid': self.api.sid} if self.api.sid else None,
-        )['nodes']
+        return self.api.get('project/nodes', params={'prjUUID': self.uuid})['nodes']
+
+    def get_report_list(self) -> List[Dict[str, str]]:
+        """Returns a list of project reports. 
+
+        .. versionadded:: 0.30.0
+        """
+        return self.api.get('project/reports', params={'prjUUID': self.uuid})
 
     def get_execution_stats(self, skip_hidden: Optional[bool] = None) -> List[Node]:
         """Returns nodes execution statistics.
@@ -61,14 +68,9 @@ class Project:
         """
         params = {'prjUUID': self.uuid}
         if skip_hidden is not None:
+            # to allow using older PA's that don't have this arg and return 'redundant parameter in query' error
             params['skipHiddenNodes'] = 'true' if skip_hidden else 'false'
-        try:
-            return self.api.get('project/execution-statistics', params=params)['nodes']
-        except APIException as exc:
-            if 'redundant parameter in query' in str(exc):
-                raise APIException(
-                    "Current PolyAnalyst server doesn't support `skip_hidden` parameter, try again without it."
-                )
+        return self.api.get('project/execution-statistics', params=params)['nodes']
 
     def get_tasks(self) -> List[Dict[str, Any]]:
         """Returns task list info."""
@@ -86,12 +88,19 @@ class Project:
         """Aborts the execution of all nodes in the project."""
         self.api.post('project/global-abort', json={'prjUUID': self.uuid})
 
-    def execute(self, *args: Union[str, Dict[str, str]], wait: bool = False) -> Optional[int]:
+    def execute(
+            self, *args: Union[str, Dict[str, str]], wait: bool = False, ignore_first_n_pabusy: int = 0
+    ) -> Optional[int]:
         """
         Initiates execution of nodes and returns execution wave identifier.
 
         :param args: node names and/or dicts with name and type of nodes
         :param wait: wait for nodes execution to complete
+        :param ignore_first_n_pabusy: (deprecated) when waiting for execution to complete skip the first specified
+            numbers of PABusy
+
+        .. versionchanged:: 0.34.0
+           Added `ignore_first_n_pabusy` parameter
 
         Usage::
 
@@ -116,16 +125,20 @@ class Project:
 
           >>> prj.execute()
         """
+
+        if ignore_first_n_pabusy:
+            warnings.warn(
+                'ignore_first_n_pabusy is deprecated, use pabusy_timeout parameter in API class instead.',
+                DeprecationWarning,
+                stacklevel=2,
+            )
+
         nodes = []
         for arg in args:
             node = self._find_node(arg)
             nodes.append({'name': node['name'], 'type': node['type']})
 
-        resp, _ = self.api.request(
-            'project/execute',
-            method='post',
-            json={'prjUUID': self.uuid, 'nodes': nodes},
-        )
+        resp, _ = self.api.request('project/execute', 'POST', json={'prjUUID': self.uuid, 'nodes': nodes})
 
         location = resp.headers.get('location')
         query = urlparse(location).query
@@ -140,11 +153,48 @@ class Project:
                     self.wait_for_completion(node)  # type: ignore
                 return
 
-            while self.is_running(wave_id):
+            while True:
                 time.sleep(1)
+                try:
+                    if not self.is_running(wave_id):
+                        break
+                except PABusy:
+                    ignore_first_n_pabusy -= 1
+                    if ignore_first_n_pabusy <= 0:
+                        raise
 
         return wave_id
 
+    def execute_to(self, node: Union[str, Dict[str, str]], wait: bool = False) -> Optional[int]:
+        """
+        Execute a sequence of nodes to the given node.
+
+        Similar to :meth:`Project.execute`, except for two differences:
+          - accepts only one node
+          - execution of a sequence of nodes stops at the given node and does not continue further
+
+        :param node: node name or a dict with node name and type
+        :param wait: wait until node is executed
+
+        :return: execution task identifier if `wait` is False else None
+
+        .. versionadded:: 0.30.0
+        """
+        resp, _ = self.api.request(
+            'project/execute-to-node',
+            method='post',
+            json={'prjUUID': self.uuid, 'nodes': [self._find_node(node)]},
+        )
+        location = resp.headers.get('location')
+        query = urlparse(location).query
+        wave_id = int(parse_qs(query).get('executionWave')[0])
+
+        if not wait:
+            return wave_id
+
+        while self.is_running(wave_id):
+            time.sleep(1)
+
     def is_running(self, wave_id: int) -> bool:
         """
         Checks that execution wave is still running in the project.
@@ -178,9 +228,88 @@ class Project:
         """
         return Parameters(self, self._find_node(name)['id'])
 
-    def unload(self) -> None:
-        """Unload the project from the memory and free system resources."""
-        self.api.post('project/unload', json={'prjUUID': self.uuid})
+    def _get_load_status(self, load_id: int) -> Dict[str, Any]:
+        """Get the status of project load.
+
+        :param load_id: the load identifier
+        :raises: APIException if bad load_id or version of PolyAnalyst older than 2815
+
+        :return: project loading status
+
+        .. versionadded:: 0.31
+        """
+
+        return self.api.get('project/load/status', params={'asyncOperationId': load_id})
+
+    def load(self, wait: bool = False) -> Union[str, Dict[str, Any]]:
+        """Load project to the memory.
+
+        :param wait: wait until the project loading is completed. Defaults to ``False``.
+        :raises: APIException if version of PolyAnalyst older than 2815
+
+        :return: load identifier if `wait` is False and load status otherwise
+
+        .. versionadded:: 0.31
+        """
+        resp, _ = self.api.request('project/load', method='post', json={'prjUUID': self.uuid})
+        location = resp.headers.get('location')
+        qs = parse_qs(urlparse(location).query)
+        load_id = qs['asyncOperationId'][0]
+
+        if not wait:
+            return load_id
+
+        while True:
+            time.sleep(1)
+            status = self._get_load_status(load_id)
+            if status.get('status') not in ('Processing'):
+                if status.get('status') in ('Error'):
+                    raise ClientException(f"Failed to load project: {status.get('message')}")
+                return status
+
+    def status(self):
+        """Get project status.
+
+        :raises: APIException if version of PolyAnalyst older than 2815
+
+        :return: project status
+
+        .. versionadded:: 0.31
+        """
+        return self.api.get('project/status', params={'prjUUID': self.uuid})
+
+    def unload(self, force_unload: bool = False) -> None:
+        """
+        Unload the project from the memory.
+
+        From version `0.26.2` this function ensures that the project is unloaded despite PABusy error
+        by repeating requests (maximum 10) until PA either returns an ok response or
+        returns an error 'the project has not been opened', which means project is also unloaded.
+
+        :raises: PABusy if PABusy returned for all 10 request attempts
+        :raises: APIException if the project has been unloaded before this function was called
+
+        .. versionchanged:: 0.30.0
+           Added `force_unload` parameter, which when used allows server to unload a project even with running nodes
+        """
+        data = {'prjUUID': self.uuid}
+        if force_unload:
+            data['forceUnload'] = True
+
+        for n in range(10):
+            try:
+                self.api.post('project/unload', json=data)
+                break
+            except PABusy:
+                time.sleep(0.1 * (2 ** (n - 1)))  # urllib's backoff formula
+            except APIException as e:
+                if 'has not been opened by user administrator' in str(e) and n != 0:
+                    warnings.warn(str(e))
+                    break
+                else:
+                    raise
+        else:
+            raise PABusy
 
     def repair(self) -> None:
         """Initiate the project repairing operation."""
@@ -255,6 +384,100 @@ class Project:
 
             time.sleep(1)
 
+    def export(
+        self, file_path: Union[str, os.PathLike], options: Optional[Dict] = None, wait: bool = False
+    ) -> Union[str, Dict[str, Any]]:
+        """
+        Export project to a server file system.
+
+        :param file_path: The path to which you want to export a project. Can be absolute path or contain PolyAnalyst
+            alias, i.e. `Open access folder` or `User home folder`. The file extension defines project format: ``ps6``,
+            ``pa6``, ``psar6``, ``paar6`` or ``pagridar6``.
+        :param options: Dict of export parameters:
+            *
+
+              | ``ids``: List of project IDs to export into one file: ``psar6``, ``paar6`` or ``pagridar6``. Defaults\
+                to current project.
+              | ``compressionLevel``: Compression level: ``store``, ``fastest``, ``fast``, ``normal``, ``maximum`` or\
+                ``ultra``. Defaults to ``normal``.
+              | ``keepBackups``: Keep backup versions. Defaults to ``True``.
+              | ``keepMacrosAndVars``: Keep user and server macros and variables. Defaults to ``True``.
+              | ``keepSliceStatistics``: Keep slice usage statistics. Defaults to ``False``.
+              | ``overwriteExisting``: Replace existing file. Defaults to ``False``.
+              | ``dicts``: Dict of project dictionaries. Defaults to ``{}``.
+              | ``usedDicts``: Dict of server-side non-embedded dictionaries. Defaults to ``{}``.
+              | ``usedParsers``: List of used parsers to include. Defaults to ``[]``.
+        :param wait: Wait until the project export is completed. Defaults to ``False``.
+        :raises ClientException: if ``compressionLevel`` is not one of the defined
+
+        :return: export identifier if `wait` is False and export status otherwise
+
+        Usage::
+
+          >>> prj.export('@administrator@/example.pa6')
+          >>> # or
+          >>> status = prj.export(
+          >>>     'C:/Users/Public/Documents/test.paar6',
+          >>>     {
+          >>>         'overwriteExisting': True,
+          >>>         'compressionLevel': 'maximum',
+          >>>     },
+          >>>     wait=True
+          >>> )
+
+        .. versionadded:: 0.28.0
+        """
+        if options is None:
+            options = {}
+
+        options['fileName'] = os.fspath(file_path)
+        options['fileFormat'] = os.fspath(file_path).split('.')[-1]
+        options['ids'] = options.get('ids', [self.uuid])
+
+        cmpr = options.get('compressionLevel')
+        compression_levels = {'store': 0, 'fastest': 1, 'fast': 3, 'normal': 5, 'maximum': 7, 'ultra': 9}
+        if cmpr:
+            try:
+                options['compressionLevel'] = compression_levels[cmpr.lower()]
+            except KeyError as e:
+                raise ClientException(f'Unknown compression level: {cmpr}') from e
+
+        defaults = {
+            'compressionLevel': 5,
+            'keepBackups': True,
+            'keepMacrosAndVars': True,
+            'keepSliceStatistics': False,
+            'overwriteExisting': False,
+            'dicts': {},
+            'usedDicts': {},
+            'usedParsers': [],
+        }
+        resp, _ = self.api.request('project/export', method='post', json={**defaults, **options})
+
+        location = resp.headers.get('location')
+        qs = parse_qs(urlparse(location).query)
+        export_id = qs['exportId'][0]
+
+        if not wait:
+            return export_id
+
+        while True:
+            time.sleep(1)
+            status = self.get_export_status(export_id)
+            # status has only empty state key when the server rebooted during the project export
+            if status.get('state') in ('Exported', 'Error'):
+                return status
+
+    def get_export_status(self, export_id: str) -> Dict[str, Any]:
+        """Get the status of project export
+
+        :param export_id: the export identifier
+
+        .. versionadded:: 0.28.0
+        """
+
+        return self.api.get('project/export/status', params={'exportId': export_id})
+
     def get_nodes(self) -> Nodes:
         """Returns a dictionary of project's nodes information.
 
@@ -283,8 +506,7 @@ class Project:
            Use :meth:`Project.get_execution_stats` instead.
         """
         warnings.warn(
-            'Project.get_execution_statistics() is deprecated, use'
-            'Project.get_execution_stats() instead.',
+            'Project.get_execution_statistics() is deprecated, use' 'Project.get_execution_stats() instead.',
             DeprecationWarning,
             stacklevel=2,
         )
@@ -320,8 +542,7 @@ class Project:
         Set parameters of the selected Parameters node in the project.
 
         :param node: name of Parameters node
-        :param node_type: node type, which parameters need to be set. The types \
-            are listed in NodeTypes.
+        :param node_type: node type, which parameters need to be set.
         :param parameters: default parameters of the node to be set.
         :param declare_unsync: reset the status of the Parameters node.
         :param hard_update: update every child node with new parameters if True, \
@@ -485,9 +706,7 @@ class DataSet:
 
         return self._api.get('dataset/preview', params=params)
 
-    def iter_rows(
-        self, start: int = 0, stop: Optional[int] = None
-    ) -> Iterator[Dict[str, JSON_VAL]]:
+    def iter_rows(self, start: int = 0, stop: Optional[int] = None) -> Iterator[Dict[str, JSON_VAL]]:
         """
         Iterate over rows in dataset.
 
@@ -513,9 +732,7 @@ class DataSet:
 
         # предполагается что если stop определен то пользователь в курсе количества строк в датасете
         if not 0 <= start <= stop <= max_row:
-            raise ValueError(
-                f'start and stop arguments must be within dataset row range: (0, {max_row})'
-            )
+            raise ValueError(f'start and stop arguments must be within dataset row range: (0, {max_row})')
 
         rows = self._values(stop)['table']
         get_text = self._cell_text
@@ -532,17 +749,18 @@ class DataSet:
                     raise StopIteration
 
                 result = {}
-                for column in info['columnsInfo']:
+                columns = info.get('columnsInfo') or info['columns']
+                for column in columns:
                     if column['flags'].get('getTextAlways'):
                         result[column['title']] = get_text(self.idx, column['id'])
                     # elif column['type'] == 'DateTime':  # todo convert to python datetime?
                     else:
                         _value = rows[self.idx][column['id']]
-                        if _value == 1e+100:
+                        if _value == 1e100:
                             _value = None
-                        elif _value == 8e+100:
+                        elif _value == 8e100:
                             _value = math.inf
-                        elif _value == -8e+100:
+                        elif _value == -8e100:
                             _value = -math.inf
 
                         result[column['title']] = _value
